#!/usr/bin/perl -w
use strict;
use warnings;
use Digest::MD5 "md5_base64";
use POSIX;
use v5.10;

$| = 0; #turn off IO buffering

# set CONFIRMATION_PHRASE to anything falsy (like zero) if you are a fast and
# loose badass and you don't need no stinkin' safety net.
use constant CONFIRMATION_PHRASE => "YES";

my $termios = POSIX::Termios->new;
$termios->getattr(fileno(STDIN));
# fetch this value once and cache it
my $startup_lflag = $termios->getlflag();

# set termios lflag to not echo keystrokes, and take us out of line mode
$termios->setlflag($startup_lflag & ~(ECHO | ECHOK | ICANON));
$termios->setattr(fileno(STDIN), TCSANOW);

END { # reset termios lflag on exit
	$termios->setlflag($startup_lflag);
	$termios->setattr(fileno(STDIN), TCSANOW);
}

if ($>) {
	say "This program requires root, sorry.";
	exit;
}

# define some colors for convenience
my %color = (
	# use tput to support anything with a terminfo entry
	"brightred" => `tput setaf 9`,
	"red" => `tput setaf 1`,
	"green" => `tput setaf 2`,
	"reset" => `tput sgr0`
);

my $notreally = grep /--notreally/, @ARGV;
my $longrun = grep /--long/, @ARGV;

if (!$notreally && CONFIRMATION_PHRASE) {
	say "$color{brightred}WARNING$color{reset}: This program will $color{brightred}COMPLETELY DESTROY ALL DATA ON ALL SYSTEM DRIVES$color{reset}";
	print 'Please type "'.CONFIRMATION_PHRASE.qq(" if you are $color{brightred}CERTAIN$color{reset} you want to continue: );
	chomp(my $confirm = <STDIN>);
	do { say; exit } unless CONFIRMATION_PHRASE eq $confirm;
}

# create our random data and store it in RAM so we can keep reusing it
open(my $drand, "<", "/dev/urandom");
my $clobber;
read($drand, $clobber, 10240000);
close($drand);
my $clobsum = md5_base64($clobber);

my @failed = ();
my @good = ();

# for providing a waiting prompt
sub anykey {
	print shift;
	return getc(STDIN);
}

# to handle errors without using die
sub warning {
	print shift;
	return 0;
}

# home made uniq function
sub uniq {
	my %seen;
	return grep { !$seen{$_}++ } @_;
}

# To get the list of devices we will be cleaning
sub getdisks {
	my $devices = `find /dev -name "sd*[a-z]"`;
	my %listhash = ();
	foreach(split("\n",$devices)) {
		my $devicepath = $_;
		my $serial = `udevadm info --query=all --name=$_ | grep SERIAL_SHORT | sed -e 's/^.*=//'`;
		my $model  = `udevadm info --query=all --name=$_ | grep "ID_MODEL="  | sed -e 's/^.*=//'`;
		chomp( $serial );
		chomp( $model );
		if ( $serial && $model ) {
			say "$devicepath\t$serial";
			$listhash{$_}{path} = $devicepath;
			$listhash{$_}{serial} = $serial;
			$listhash{$_}{model} = $model;
		} else { next; }
	}
	return %listhash;
}

# To whipe a drive
sub wipethemdrives {
	my $diskid = shift;
	my $disks = shift;
	# Write
	say "$disks->{$diskid}{path} - Serial:$disks->{$diskid}{serial} is being wiped...";
	my $disksum;
	my $disksum2;
	my $spacer = 29;
	if ($notreally) {
		$spacer = 18;
		print "Pretend wipe complete, pretending to check...";
		$disksum = $clobsum;
		$disksum2 = $clobsum;
	} else {
		open(my $diskw, ">", $disks->{$diskid}{path}) or return warning("could not open $disks->{$diskid}{path}");
		print $diskw $clobber;
		seek($diskw,-10240000,2);
		print $diskw $clobber;
		close($diskw);
		system("sync");
		print "Wipe attempt complete, checking...";
		# Read
		open(my $diskr, "<", $disks->{$diskid}{path}) or return warning("could not open $disks->{$diskid}{path}");
		my $diskdata;
		read($diskr, $diskdata, 10240000);
		$disksum = md5_base64($diskdata);
		seek($diskr,-10240000,2);
		read($diskr, $diskdata, 10240000);
		$disksum2 = md5_base64($diskdata);
		close($diskr);
	}
	if ($disksum eq $clobsum) {
		say +("."x$spacer)."Checksum $color{green}PASSED$color{reset}";
		print "             Checking raid wipe...";
		if ($disksum2 eq $clobsum) {
			say +("."x$spacer)."Checksum $color{green}PASSED$color{reset}";	
			return 1;
		} else {
			say +("."x$spacer)."Checksum $color{red}FAILED$color{reset}";
			return 0;
		}
	} else {
		say +("."x$spacer)."Checksum $color{red}FAILED$color{reset}";
		return 0;
	}
#	If long flag than zero whole disk and give a progress bar
	if ($longrun && !$notreally) {
		say "Zeroing $disks->{$diskid}{path}...";
		open(my $longdiskw, ">:raw", $disks->{$diskid}{path}) or return warning("could not open $disks->{$diskid}{path}");
		seek($longdiskw,0,2);
		my $totalBytes = tell($longdiskw);
		seek($longdiskw,0,0);
		until (eof $longdiskw) {
			print $longdiskw "\0";
        	if ( !(tell($lingdiskw) % ($totoalBytes / 1000) ) {
				printf(" %5.5s%%\r\r\r\r\r\r\r", sprintf("%.2f", (tell($longdiskw)/$totalBytes)) );
            }
		}
	}
}

# to test a drive
sub smartcheck {
# Do our smart disk thing...
	my $diskid = shift;
	my $disks = shift;
	print "Checking overall-health self-assessment test result: ".(" "x19);
	(my $smartassess = `smartctl /dev/sda -H | grep overall`) or (say "$!\ncould not check smart data, skipping" and return 1);
	$disks->{$diskid}{selftest} = (split(": ", $smartassess))[1];
	if ( "FAILED" eq $disks->{$diskid}{selftest} ) { say "$color{red}Failed$color{reset}"; return 0 
	} else { say "$color{green}Passed$color{reset}"; }
	say "Checking S.M.A.R.T. variables...";
#	say qq(smartctl $disks->{$diskid}{path} -A -f hex,id | grep "^0x");
	(my $smartcommand = `smartctl $disks->{$diskid}{path} -A -f hex,id | grep "^0x"`) or (say "$!\ncould not check smart data, skipping" and return 1);
	for (split("\n",$smartcommand)) {
		my @smartdata = split(/\s+/,$_);
		$disks->{$diskid}{smart}{$smartdata[0]}{id} = $smartdata[0];
		$disks->{$diskid}{smart}{$smartdata[0]}{name} = $smartdata[1];
		$disks->{$diskid}{smart}{$smartdata[0]}{flag} = $smartdata[2];
		$disks->{$diskid}{smart}{$smartdata[0]}{value} = $smartdata[3];
		$disks->{$diskid}{smart}{$smartdata[0]}{worst} = $smartdata[4];
		$disks->{$diskid}{smart}{$smartdata[0]}{thresh} = $smartdata[5];
		$disks->{$diskid}{smart}{$smartdata[0]}{type} = $smartdata[6];
		$disks->{$diskid}{smart}{$smartdata[0]}{updated} = $smartdata[7];
		$disks->{$diskid}{smart}{$smartdata[0]}{when_failed} = $smartdata[8];
		$disks->{$diskid}{smart}{$smartdata[0]}{raw_value} = $smartdata[9];
	}
	close( $smartcommand );
# these are incredibly strict requirements, you probably should be looking at the when_failed field instead...
	for my $smartentry (sort {$a cmp $b} keys %{$disks->{$diskid}{smart}}) {
		if ($disks->{$diskid}{smart}{$smartentry}{id} =~ /0x05|0x07|0xc4|0xc5|0xc6|0xc7/) {
			printf("%-6.6s %-26.26s\t%-26.26s\t", $disks->{$diskid}{smart}{$smartentry}{id}, $disks->{$diskid}{smart}{$smartentry}{name}, $disks->{$diskid}{smart}{$smartentry}{raw_value});
			if ($disks->{$diskid}{smart}{$smartentry}{raw_value} =~ /0|0\/0/) {
				say "$color{green}Passed$color{reset}";
			} else { say "$color{red}Failed$color{reset}"; return 0}
			
		}
	}
	return 1;
}

# main loop
while () {
	my %disks = getdisks();
	my $batchcount = keys %disks;
	say "$color{red}WARNING:$color{reset} This program will completely destroy all data on $batchcount drives with the following serial numbers:";
	foreach my $diskid (sort {$disks{$a}{serial} cmp $disks{$b}{serial}} keys %disks) {
		say $disks{$diskid}{serial};
	}
#	exit unless "\n" eq anykey("Last chance to save your data, hit ENTER to DESTROY ALL DATA, or any other key to abort!\n");
	foreach my $diskid (sort {$a cmp $b}keys %disks) {
		print "="x33;
		printf '[ %-10s ]', $diskid;
		say "="x33;
		$disks{$diskid}{wipe} = wipethemdrives($diskid,\%disks);
		if ($disks{$diskid}{wipe} == 1) {
			$disks{$diskid}{smartpass} = smartcheck($diskid,\%disks);
			if ($disks{$diskid}{smartpass} == 1) {
				push(@good, sprintf("%-26.26s\t%46.46s",$disks{$diskid}{serial},$disks{$diskid}{model}));
			} else { push(@failed, sprintf("%-26.26s\t%46.46s",$disks{$diskid}{serial},$disks{$diskid}{model})); }
		} else { push(@failed, $disks{$diskid}{serial}); }
		print "="x80;
		print "\n"x5;
	}
	say "These drives were successfully wiped and passed SMART, they may be put back into\nproduction:";
	print $color{green}.join("\n",sort {$a cmp $b} uniq(@good)).$color{reset};
	anykey("\nHit any key to see the list of failed drives. Please also note any drives that\ndo not appear on either list. Those should be tried again (if they do not show\nup after three times assume failure).\n");
	say "$color{red}The following drives are bad, RMA or shred:";
	say join("\n",sort {$a cmp $b} uniq(@failed));
	print "$color{green}";
	@failed = ();
	@good = ();
	say $color{green}.("-"x34).$color{red}."\nWipe complete, insert more drives.\n".$color{green}.("-"x34).$color{reset};
	anykey("Hit any key when ready to continue...\n");
}
